#!BPY
""" Registration info for Blender menus:
Name: 'Jake Format (.jfm)...'
Blender: 235
Group: 'Export'
Submenu: 'All Objects...' all
Submenu: 'Selected Objects...' selected
Tooltip: 'Export to Jake Format (.jfm)'
"""

__author__ = ("Rick Kimball", "Ken Miller", "Steve Matthews", "Bart")
__url__ = ["blender", "elysiun",
"Author's (Rick) homepage, http://kimballsoftware.com/blender",
"Author's (Bart) homepage, http://www.neeneenee.de/vrml"]
__email__ = ["Bart, bart:neeneenee*de"]
__version__ = "2006/01/17"
__bpydoc__ = """\
This script exports to Jake format.

Usage:

Run this script from "File->Export" menu.  A pop-up will ask whether you
want to export only selected or all relevant objects.

Known issues:<br>
    Doesn't handle multiple UV textures on a single mesh (create a mesh
for each texture);<br>
    Can't get the texture array associated with material * not the UV ones;
"""


# $Id: jakeformat_export.py,v 1.7 2006/01/17 22:13:25 campbellbarton Exp $
#
#------------------------------------------------------------------------
# Jake Format exporter for blender 2.36 or above
#
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
#

####################################
# Library dependancies
####################################

import Blender, cStringIO, struct
from Blender import Object, NMesh, Draw, BGL, Image, Text, sys, Mathutils, Mesh
from Blender.Scene import Render
try:
    from os.path import exists, join
    pytinst = 1
except:
    print "No Python installed, for full features install Python (http://www.python.org/)."
    pytinst = 0
import math

####################################
# Global Variables
####################################

scene = Blender.Scene.getCurrent()
world = Blender.World.Get() 
worldmat = Blender.Texture.Get()
filename = Blender.Get('filename')
_safeOverwrite = True
extension = ''
ARG=''

class DrawTypes:
    """Object DrawTypes enum values
    BOUNDS - draw only the bounding box of the object
    WIRE - draw object as a wire frame
    SOLID - draw object with flat shading
    SHADED - draw object with OpenGL shading
"""
    BOUNDBOX  = 1
    WIRE      = 2
    SOLID     = 3
    SHADED    = 4
    TEXTURE   = 5

if not hasattr(Blender.Object,'DrawTypes'):
    Blender.Object.DrawTypes = DrawTypes()

##########################################################
# Functions for writing output file
##########################################################

class JFM2Export:

    def __init__(self, filename):
        #--- public you can change these ---
	self.currentindex = 0
        self.wire = 0
        self.proto = 1
        self.matonly = 0
        self.share = 0
        self.verbose=2     # level of verbosity in console 0-none, 1-some, 2-most
        self.cp=3          # decimals for material color values     0.000 - 1.000
        self.vp=3          # decimals for vertex coordinate values  0.000 - n.000
        self.tp=3          # decimals for texture coordinate values 0.000 - 1.000
        self.it=3
        
        #--- class private don't touch ---
        self.texNames={}   # dictionary of textureNames
        self.filename=filename
        self.file = open(filename, "w")
        self.bNav=0
        self.nodeID=0

##########################################################
# Writing nodes routines
##########################################################

    def writeIndexedFaceSet(self, object, normals = 0):
	data = cStringIO.StringIO()

        vColors={}    # 'multi':1
        mesh=object.getData()

        if len(mesh.faces) == 0:
					return
        for face in mesh.faces:
            if face.mode & Blender.NMesh.FaceModes['OBCOL'] and self.matonly == 0:
                self.matonly = 1
            elif face.mode & Blender.NMesh.FaceModes['SHAREDCOL'] and self.share == 0:
                self.share = 1
                
        om = object.getMatrix();
            
        maters=mesh.materials
        hasImageTexture=0
        issmooth=0

        if len(maters) > 0 or mesh.hasFaceUV():
          # right now this script can only handle a single material per mesh.
          if len(maters) >= 1:
        
            #-- textures
            if mesh.hasFaceUV():
                for face in mesh.faces:
                    if (hasImageTexture == 0) and (face.image):
                        hasImageTexture=1  # keep track of face texture

        #-- IndexedFaceSet or IndexedLineSet

        # check if object is wireframe only
        if object.drawType == Blender.Object.DrawTypes.WIRE:
            # user selected WIRE=2 on the Drawtype=Wire on (F9) Edit page
            ifStyle="IndexedLineSet"
            self.wire = 1
        else:
            # user selected BOUNDS=1, SOLID=3, SHARED=4, or TEXTURE=5
            ifStyle="IndexedFaceSet"

        #--- output coordinates
	self.writeNumCoordinates(data, object, mesh)
	self.writeNumVertInd(data, object, mesh)
	self.writeNumTextureCoordinates(data, mesh)
	self.writeNumTextureCoordinateInd(data, mesh)
	self.writeNumFrames(data)

	for framenum in range(Blender.Get('staframe'), Blender.Get('endframe') + 1):		
		Blender.Set('curframe', framenum)
	        self.writeCoordinateVerts(data, object)

	#self.writeCoordinateIndices(data, object, mesh)

        if object.drawType != Blender.Object.DrawTypes.WIRE:
            #--- output textureCoordinates if UV texture used
            if mesh.hasFaceUV():
                if hasImageTexture == 1:
                    self.writeTextureCoordinates(data, mesh)
                elif self.matonly == 1 and self.share == 1:
                    self.writeFaceColors(mesh)

        #--- output vertexColors
        if self.share == 1 and self.matonly == 0:
            self.writeVertexColors(mesh)

        self.matonly = 0
        self.share = 0
        self.wire = 0

	return data.getvalue()	

    def writeNumCoordinates(self, data, object, mesh):
	nvert = 0
	meshVertexList = mesh.verts
	for vertex in meshVertexList:
		nvert = nvert + 3    # 3 coordinates for each vertex
	data.write(struct.pack("I", nvert))

    def GetNV(self, mesh):
	nvert = 0
	meshVertexList = mesh.verts
	for vertex in meshVertexList:
		nvert = nvert + 3    # 3 coordinates for each vertex
	return nvert	

    def writeNumFrames(self, data):
	nframes = (Blender.Get('endframe') - Blender.Get('staframe')) + 1
	#nframes = 1

	data.write(struct.pack("I", nframes))

    def writeNumVertInd(self, data, object, mesh):
	nind = 0
	for face in mesh.faces:
		for i in range(len(face)):
			nind = nind + 1
		nind = nind + 1   # Add one for the -1 value that finishes each face
	data.write(struct.pack("I", nind))

    def GetNVI(self, mesh):
	nind = 0
	for face in mesh.faces:
		for vert in face.v:
			nind = nind + 1
		nind = nind + 1   # Add one for the -1 value that finishes each face
	return nind

    def writeCoordinateVerts(self, data, object):
	mesh = object.getData(False, True)
	mesh.getFromObject(object.name)
	meshVertexList = mesh.verts
	
        # create vertex list
	#vertex = meshVertexList[0].co
	#print "%s %s: %s %s %s" % (object.name, vertex.co[0], vertex.co[1], vertex.co[2])

        for vertex in meshVertexList:
		mm=object.getMatrix()
		v=self.rotVertex(mm, vertex.co)
		data.write(struct.pack("fff", v[0], v[1], v[2]))
		#print vertex
	
    def writeCoordinateIndices(self, data, object, mesh):
	meshVertexList = mesh.verts
        coordIndexList=[]  
        for face in mesh.faces:

        	for i in range(len(face)):
                	indx=meshVertexList.index(face[i]) + (self.currentindex)
			data.write(struct.pack("I", indx))
			print "range: %s index: %s "  %   (range(len(face)), indx)
		data.write(struct.pack("I", -1))		
		print "------------ next face"

    def writeNumTextureCoordinates(self, data, mesh):
	nTC = 0
        texCoordList=[] 
        texIndexList=[]
        j=0

        for face in mesh.faces:
            for i in range(len(face)):
                texIndexList.append(j)
                texCoordList.append(face.uv[i])
                j=j+1
            texIndexList.append(-1)

        for i in range(len(texCoordList)):
		nTC = nTC + 2
	data.write(struct.pack("I", nTC))

    def GetNT(self, mesh):
	nTC = 0
        texCoordList=[] 
        texIndexList=[]
        j=0

        for face in mesh.faces:
#	    nTC = (len(face) + 1) * 2
            for i in range(len(face)):
                texIndexList.append(j)
                texCoordList.append(face.uv[i])
                j=j+1
            texIndexList.append(-1)

        for i in range(len(texCoordList)):
		nTC = nTC + 2
	return nTC

    def writeNumTextureCoordinateInd(self, data, mesh):
	nTCI = 0
        for face in mesh.faces:
            for i in range(len(face)):
		nTCI = nTCI + 1
	    nTCI = nTCI + 1    # Add one for the extra -1 added to the list
	data.write(struct.pack("I", nTCI))

    def GetNTI(self, mesh):
	nTCI = 0
        for face in mesh.faces:
            for i in range(len(face)):
		nTCI = nTCI + 1
	    nTCI = nTCI + 1    # Add one for the extra -1 added to the list
	return nTCI

    def writeTextureCoordinates(self, data, mesh):
        texCoordList=[] 
        texIndexList=[]
        j=0

        for face in mesh.faces:
            for i in range(len(face)):
                texIndexList.append(j)
                texCoordList.append(face.uv[i])
                j=j+1
            texIndexList.append(-1)

        for i in range(len(texCoordList)):
	    data.write(struct.pack("ff", texCoordList[i][0], texCoordList[i][1]))

    def writeTextureCoordinateIndices(self, data, mesh):
        texCoordList=[] 
        texIndexList=[]
        j=0

        for face in mesh.faces:
            for i in range(len(face)):
                texIndexList.append(j)
                texCoordList.append(face.uv[i])
                j=j+1
            texIndexList.append(-1)

        texIndxStr=""
        for i in range(len(texIndexList)):
#            texIndxStr = texIndxStr + "%d\n " % texIndexList[i]
   	    data.write(struct.pack("I", texIndexList[i]  + (self.currentindex)))

            if texIndexList[i]==-1:
                texIndxStr=""

    def writeFaceColors(self, mesh):
        for face in mesh.faces:
            if face.col:
                c=face.col[0]
                if self.verbose > 2:
                    print "Debug: face.col r=%d g=%d b=%d" % (c.r, c.g, c.b)

                aColor = self.rgbToFS(c)

    def writeVertexColors(self, mesh):
        for i in range(len(mesh.verts)):
            c=self.getVertexColorByIndx(mesh,i)
            if self.verbose > 2:
                print "Debug: vertex[%d].col r=%d g=%d b=%d" % (i, c.r, c.g, c.b)

            aColor = self.rgbToFS(c)

    def rotVertex(self, mm, v):
        lx,ly,lz=v[0],v[1],v[2]
        gx=(mm[0][0]*lx + mm[1][0]*ly + mm[2][0]*lz) + mm[3][0]
        gy=((mm[0][2]*lx + mm[1][2]*ly+ mm[2][2]*lz) + mm[3][2])
        gz=-((mm[0][1]*lx + mm[1][1]*ly + mm[2][1]*lz) + mm[3][1])
        rotatedv=[gx,gy,gz]
        return rotatedv

##########################################################
# export routine
##########################################################

    def export(self, scene, world, worldmat):
	print "Info: starting Jake Format export to " + self.filename + "..."

	nv = 0;  # num vertices
	nvi = 0; # num vertex indices
	nt = 0;  # num texture coords
	nti = 0; # num texture coord indices
	
	data = cStringIO.StringIO()
	self.proto = 0
        allObj = []
        if ARG == 'selected':
            allObj = Blender.Object.GetSelected()
        else:
            allObj = scene.getChildren()

        for thisObj in allObj:
            try:
                objType=thisObj.getType()
                objName=thisObj.getName()
    		mesh=thisObj.getData()
            	self.matonly = 0
                if objType == "Mesh":
		    nv = nv + self.GetNV(mesh)
		    nvi = nvi + self.GetNVI(mesh)
		    nt = nt + self.GetNT(mesh)
		    nti = nti + self.GetNTI(mesh)	
                else:
                    print ""
            except AttributeError:
                print "Error: Unable to get type info for %s" % thisObj.getName()

        print("%s %s %s %s" % (nv, nvi, nt, nti))

	data.write(struct.pack("I", nv))
	data.write(struct.pack("I", nvi))
	data.write(struct.pack("I", nt))
	data.write(struct.pack("I", nti))

	self.writeNumFrames(data)
	self.currentindex = 0

	# Write out verts
	for framenum in range(Blender.Get('staframe'), Blender.Get('endframe') + 1):		
		for thisObj in allObj:
			objType=thisObj.getType()
                	if objType == "Mesh":
				Blender.Set('curframe', framenum)
	        		self.writeCoordinateVerts(data, thisObj)

	# Write out vert indices
	for thisObj in allObj:
		objType=thisObj.getType()
                if objType == "Mesh":
			print ("Vert indices for: %s" % thisObj.name)
			mesh = thisObj.getData()
			self.writeCoordinateIndices(data, thisObj, mesh)
			# Start vert_indices after end of previous object's verts
			self.currentindex = (self.currentindex + (self.GetNV(mesh)/3))
			print ("currentindex: %s " % self.currentindex)

	# reset currentindex for use with texture coordinates
	self.currentindex = 0

	# Write out texture coords
	for thisObj in allObj:
		objType=thisObj.getType()
               	if objType == "Mesh":
			mesh=thisObj.getData()			
			self.writeTextureCoordinates(data, mesh)

	# Write out texture coords
	for thisObj in allObj:
		objType=thisObj.getType()
               	if objType == "Mesh":
			mesh=thisObj.getData()
			self.writeTextureCoordinateIndices(data, mesh)
			self.currentindex = (self.currentindex + (self.GetNT(mesh)/2))

	file = open(self.filename, "wb")
	file.write(data.getvalue())

        self.cleanup()
        
##########################################################
# Utility methods
##########################################################

    def cleanup(self):
        self.file.close()
        self.texNames={}
        print "Info: finished Jake Format export to %s\n" % self.filename

##########################################################
# Callbacks, needed before Main
##########################################################

def select_file(filename):
  if pytinst == 1:
    if exists(filename) and _safeOverwrite:
      result = Draw.PupMenu("File Already Exists, Overwrite?%t|Yes%x1|No%x0")
      if(result != 1):
        return

  if not filename.endswith(extension):
    filename += extension

  wrlexport=JFM2Export(filename)
  wrlexport.export(scene, world, worldmat)

def createWRLPath():
  filename = Blender.Get('filename')
  print filename
  
  if filename.find('.') != -1:
    filename = filename.split('.')[0]
    filename += extension
    print filename

  return filename

#########################################################
# main routine
#########################################################

try:
    ARG = __script__['arg'] # user selected argument
except:
    print "older version"

if Blender.Get('version') < 235:
  print "Warning: Jake Format export failed, wrong blender version!"
  print " You aren't running blender version 2.35 or greater"
  print " download a newer version from http://blender3d.org/"
else:
  extension=".jfm"
  Blender.Window.FileSelector(select_file,"Export JFM",createWRLPath())
