/*
 * pcx.c -- pcx texture loader
 * last modification: dec. 28, 2005
 *
 * Copyright (c) 2005 David HENRY
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * gcc -Wall -ansi -L/usr/X11R6/lib -lGL -lGLU -lglut pcx.c -o pcx
 */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* OpenGL texture info */
typedef struct
{
  GLsizei width;
  GLsizei height;

  GLenum format;
  GLint internalFormat;
  GLuint id;

  GLubyte *texels;

} gl_texture_t;


#pragma pack(1)
/* pcx header */
typedef struct
{
  GLubyte manufacturer;
  GLubyte version;
  GLubyte encoding;
  GLubyte bitsPerPixel;

  GLushort xmin, ymin;
  GLushort xmax, ymax;
  GLushort horzRes, vertRes;

  GLubyte palette[48];
  GLubyte reserved;
  GLubyte numColorPlanes;

  GLushort bytesPerScanLine;
  GLushort paletteType;
  GLushort horzSize, vertSize;

  GLubyte padding[54];

} pcx_header_t;
#pragma pack(4)


/* texture Id */
GLuint texId;


void
ReadPCX1bit (FILE *fp, pcx_header_t *hdr, gl_texture_t *texinfo)
{
  int y, i, bytes;
  int colorIndex;
  int rle_count = 0, rle_value = 0;
  GLubyte *ptr = texinfo->texels;

  for (y = 0; y < texinfo->height; ++y)
    {
      ptr = &texinfo->texels[(texinfo->height - (y + 1)) * texinfo->width * 3];
      bytes = hdr->bytesPerScanLine;

      /* decode line number y */
      while (bytes--)
	{
	  if (rle_count == 0)
	    {
	      if ( (rle_value = fgetc (fp)) < 0xc0)
		{
		  rle_count = 1;
		}
	      else
		{
		  rle_count = rle_value - 0xc0;
		  rle_value = fgetc (fp);
		}
	    }

	  rle_count--;

	  for (i = 7; i >= 0; --i, ptr += 3)
	    {
	      colorIndex = ((rle_value & (1 << i)) > 0);

	      ptr[0] = hdr->palette[colorIndex * 3 + 0];
	      ptr[1] = hdr->palette[colorIndex * 3 + 1];
	      ptr[2] = hdr->palette[colorIndex * 3 + 2];
	    }
	}
    }
}


void
ReadPCX4bits (FILE *fp, pcx_header_t *hdr, gl_texture_t *texinfo)
{
  GLubyte *colorIndex, *line;
  GLubyte *pLine, *ptr;
  int rle_count = 0, rle_value = 0;
  int x, y, c;
  int bytes;

  colorIndex = (GLubyte *)malloc (sizeof (GLubyte) * texinfo->width);
  line = (GLubyte *)malloc (sizeof (GLubyte) * hdr->bytesPerScanLine);

  for (y = 0; y < texinfo->height; ++y)
    {
      ptr = &texinfo->texels[(texinfo->height - (y + 1)) * texinfo->width * 3];

      memset (colorIndex, 0, texinfo->width * sizeof (GLubyte));

      for (c = 0; c < 4; ++c)
	{
	  pLine = line;
	  bytes = hdr->bytesPerScanLine;

	  /* decode line number y */
	  while (bytes--)
	    {
	      if (rle_count == 0)
		{
		  if ( (rle_value = fgetc (fp)) < 0xc0)
		    {
		      rle_count = 1;
		    }
		  else
		    {
		      rle_count = rle_value - 0xc0;
		      rle_value = fgetc (fp);
		    }
		}

	      rle_count--;
	      *(pLine++) = rle_value;
	    }

	  /* compute line's color indexes */
	  for (x = 0; x < texinfo->width; ++x)
	    {
	      if (line[x / 8] & (128 >> (x % 8)))
		colorIndex[x] += (1 << c);
	    }
	}

      /* decode scan line.  color index => rgb  */
      for (x = 0; x < texinfo->width; ++x, ptr += 3)
	{
	  ptr[0] = hdr->palette[colorIndex[x] * 3 + 0];
	  ptr[1] = hdr->palette[colorIndex[x] * 3 + 1];
	  ptr[2] = hdr->palette[colorIndex[x] * 3 + 2];
	}
    }

  /* release memory */
  free (colorIndex);
  free (line);
}


void
ReadPCX8bits (FILE *fp, pcx_header_t *hdr, gl_texture_t *texinfo)
{
  int rle_count = 0, rle_value = 0;
  GLubyte palette[768];
  GLubyte magic;
  GLubyte *ptr;
  fpos_t curpos;
  int y, bytes;
  int i;

  /* the palette is contained in the last 769 bytes of the file */
  fgetpos (fp, &curpos);
  fseek (fp, -769, SEEK_END);
  magic = fgetc (fp);

  /* first byte must be equal to 0x0c (12) */
  if (magic != 0x0c)
    {
      fprintf (stderr, "error: colormap's first byte must be 0x0c! "
	       "(%#x)\n", magic);

      free (texinfo->texels);
      texinfo->texels = NULL;
      return;
    }

  /* read palette */
  fread (palette, sizeof (GLubyte), 768, fp);
  fsetpos (fp, &curpos);

  /* read pixel data */
  for (y = 0; y < texinfo->height; ++y)
    {
      bytes = hdr->bytesPerScanLine;
      //ptr = &texinfo->texels[(texinfo->height - (y + 1)) * texinfo->width * 3];
      // jake's modifications:
      ptr = &texinfo->texels[(texinfo->height - (y + 1)) * texinfo->width * 3];
      //ptr += 3*bytes; //jakemod1

      /* decode line number y */
      while (bytes--)
	{
	  if (rle_count == 0)
	    {
	      if( (rle_value = fgetc (fp)) < 0xc0)
		{
		  rle_count = 1;
		}
	      else
		{
		  rle_count = rle_value - 0xc0;
		  rle_value = fgetc (fp);
		}
	    }

	  rle_count--;

	  ptr[0] = palette[rle_value * 3 + 0];
	  ptr[1] = palette[rle_value * 3 + 1];
	  ptr[2] = palette[rle_value * 3 + 2];
	  ptr += 3;
	  // ptr -= 3; jakemod1
	}
    }
}


void
ReadPCX24bits (FILE *fp, pcx_header_t *hdr, gl_texture_t *texinfo)
{
  GLubyte *ptr = texinfo->texels;
  int rle_count = 0, rle_value = 0;
  int y, c;
  int bytes;

  for (y = 0; y < texinfo->height; ++y)
    {
      /* for each color plane */
      for (c = 0; c < 3; ++c)
	{
	  ptr = &texinfo->texels[(texinfo->height - (y + 1)) * texinfo->width * 3];
	  bytes = hdr->bytesPerScanLine;

	  /* decode line number y */
	  while (bytes--)
	    {
	      if (rle_count == 0)
		{
		  if( (rle_value = fgetc (fp)) < 0xc0)
		    {
		      rle_count = 1;
		    }
		  else
		    {
		      rle_count = rle_value - 0xc0;
		      rle_value = fgetc (fp);
		    }
		}

	      rle_count--;
	      ptr[c] = (GLubyte)rle_value;
	      ptr += 3;
	    }
	}
    }
}


gl_texture_t *
ReadPCXFile (const char *filename)
{
  gl_texture_t *texinfo;
  pcx_header_t header;
  FILE *fp = NULL;
  int bitcount;

  /* open image file */
  fp = fopen (filename, "rb");
  if (!fp)
    {
      fprintf (stderr, "error: couldn't open \"%s\"!\n", filename);
      return NULL;
    }

  /* read header file */
  fread (&header, sizeof (pcx_header_t), 1, fp);
  if (header.manufacturer != 0x0a)
    {
      fprintf (stderr, "error: bad version number! (%i)\n",
	       header.manufacturer);
      return NULL;
    }

  /* initialize texture parameters */
  texinfo = (gl_texture_t *)malloc (sizeof (gl_texture_t));
  texinfo->width = header.xmax - header.xmin + 1;
  texinfo->height = header.ymax - header.ymin + 1;
  texinfo->format = GL_RGB;
  texinfo->internalFormat = 3;
  texinfo->texels = (GLubyte *)malloc (sizeof (GLubyte) * texinfo->width
			       * texinfo->height * texinfo->internalFormat);

  bitcount = header.bitsPerPixel * header.numColorPlanes;

  /* read image data */
  switch (bitcount)
    {
    case 1:
      /* 1 bit color index */
      ReadPCX1bit (fp, &header, texinfo);
      break;

    case 4:
      /* 4 bits color index */
      ReadPCX4bits (fp, &header, texinfo);
      break;

    case 8:
      /* 8 bits color index */
      ReadPCX8bits (fp, &header, texinfo);
      break;

    case 24:
      /* 24 bits */
      ReadPCX24bits (fp, &header, texinfo);
      break;

    default:
      /* unsupported */
      fprintf (stderr, "error: unknown %i bitcount pcx files\n", bitcount);
      free (texinfo->texels);
      free (texinfo);
      texinfo = NULL;
      break;
    }

  fclose (fp);
  return texinfo;
}


void
loadtxr (const char *filename, GLuint *txr)
{
  gl_texture_t *pcx_tex = NULL;
  GLuint tex_id = 0;

  pcx_tex = ReadPCXFile (filename);

  if (pcx_tex && pcx_tex->texels)
    {
      /* generate texture */
      glGenTextures (1, &pcx_tex->id);
      glBindTexture (GL_TEXTURE_2D, pcx_tex->id);

      /* setup some parameters for texture filters and mipmapping */
      glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);


      /*
      glTexImage2D (GL_TEXTURE_2D, 0, pcx_tex->internalFormat,
		    pcx_tex->width, pcx_tex->height, 0, pcx_tex->format,
		    GL_UNSIGNED_BYTE, pcx_tex->texels);
      */

      gluBuild2DMipmaps (GL_TEXTURE_2D, pcx_tex->internalFormat,
			 pcx_tex->width, pcx_tex->height,
			 pcx_tex->format, GL_UNSIGNED_BYTE, pcx_tex->texels);

      tex_id = pcx_tex->id;

      /* OpenGL has its own copy of texture data */
      free (pcx_tex->texels);
      free (pcx_tex);
    }
  
  *txr = tex_id;
  return;// tex_id;
}
